# -*- coding: utf-8 -*-

from tempfile import TemporaryFile

from .base import Utils
from .task import IntervalTask

from hagworm.extend.base import ContextManager


class QueueBuffer:

    def __init__(self, maxsize, timeout=0):

        self._maxsize = maxsize

        self._timer = IntervalTask.create(timeout, False, self._handle_buffer) if timeout > 0 else None

        self._data_list = []

    def _handle_buffer(self):

        if len(self._data_list) > 0:
            data_list, self._data_list = self._data_list, []
            Utils.call_soon(self._run, data_list)

    async def _run(self, data_list):
        raise NotImplementedError()

    def append(self, data):

        self._data_list.append(data)

        if len(self._data_list) >= self._maxsize:
            self._handle_buffer()

    def extend(self, data_list):

        self._data_list.extend(data_list)

        if len(self._data_list) >= self._maxsize:
            self._handle_buffer()


class FileBuffer(ContextManager):
    """文件缓存类
    """

    def __init__(self, slice_size=0x1000000):

        self._buffers = []

        self._slice_size = slice_size

        self._read_offset = 0

        self._append_buffer()

    def _context_release(self):

        self.close()

    def _append_buffer(self):

        self._buffers.append(TemporaryFile())

    def close(self):

        while len(self._buffers) > 0:
            self._buffers.pop(0).close()

        self._read_offset = 0

    def write(self, data):

        buffer = self._buffers[-1]

        buffer.seek(0, 2)
        buffer.write(data)

        if buffer.tell() >= self._slice_size:
            buffer.flush()
            self._append_buffer()

    def read(self, size=None):

        buffer = self._buffers[0]

        buffer.seek(self._read_offset, 0)

        result = buffer.read(size)

        if len(result) == 0 and len(self._buffers) > 1:
            self._buffers.pop(0).close()
            self._read_offset = 0
        else:
            self._read_offset = buffer.tell()

        return result



# -*- coding: utf-8 -*-

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.extend.asyncio.base import Launcher, Utils


async def main():

    Utils.log.info(r'waiting...')

    await Utils.sleep(5)

    Utils.log.info(r'finished')


if __name__ == r'__main__':

    Launcher().run(main())



# -*- coding: utf-8 -*-

import os

from terminal_table import Table

from hagworm.extend.interface import RunnableInterface
from hagworm.extend.asyncio.base import Launcher as _Launcher
from hagworm.extend.asyncio.base import Utils, MultiTasks, AsyncCirculatorForSecond
from hagworm.extend.asyncio.zmq import Subscriber, PublisherWithBuffer


SIGNAL_PROTOCOL = r'tcp'
SIGNAL_PORT = 0x310
HIGH_WATER_MARK = 0xffffff


class Guardian(RunnableInterface):

    async def _do_polling(self, pids, hwm):

        with Reporter() as reporter:

            reporter.set_hwm(hwm)

            async for _ in AsyncCirculatorForSecond():

                for pid in pids.copy():
                    if os.waitpid(pid, os.WNOHANG)[0] == pid:
                        pids.remove(pid)

                if len(pids) == 0:
                    break

            Utils.log.info(f'\n{reporter.get_report_table()}')

    def run(self, pids):

        global HIGH_WATER_MARK

        Utils.run_until_complete(self._do_polling(pids, HIGH_WATER_MARK))


class Launcher(_Launcher):

    def __init__(self,
                 log_file_path=None, log_level=r'INFO', log_file_num_backups=7,
                 process_number=1, process_guardian=None,
                 debug=False
                 ):

        if process_guardian is None:
            process_guardian = Guardian().run

        super().__init__(
            log_file_path, log_level, log_file_num_backups,
            process_number, process_guardian,
            debug
        )

    def run(self, func, *args, **kwargs):

        if self._process_number > 1:

            super().run(func, *args, **kwargs)

        else:

            async def _func():

                nonlocal func, args, kwargs

                with Reporter() as reporter:
                    await func(*args, **kwargs)
                    Utils.log.info(f'\n{reporter.get_report_table()}')

            super().run(_func)


class Reporter(Subscriber):

    class _Report:

        def __init__(self):
            self.success = []
            self.failure = []

    def __init__(self):

        global SIGNAL_PROTOCOL, SIGNAL_PORT, HIGH_WATER_MARK

        super().__init__(f'{SIGNAL_PROTOCOL}://*:{SIGNAL_PORT}', True)

        self.set_hwm(HIGH_WATER_MARK)

        self._reports = {}

    async def _message_handler(self, data):

        for name, result, resp_time in data:
            if name and result in (r'success', r'failure'):
                getattr(self._get_report(name), result).append(resp_time)

    def _get_report(self, name: str) -> _Report:

        if name not in self._reports:
            self._reports[name] = self._Report()

        return self._reports[name]

    def get_report_table(self) -> str:

        reports = []

        for key, val in self._reports.items():
            reports.append(
                (
                    key,
                    len(val.success),
                    len(val.failure),
                    r'{:.2%}'.format(len(val.success) / (len(val.success) + len(val.failure))),
                    r'{:.3f}s'.format(sum(val.success) / len(val.success) if len(val.success) > 0 else 0),
                    r'{:.3f}s'.format(min(val.success) if len(val.success) > 0 else 0),
                    r'{:.3f}s'.format(max(val.success) if len(val.success) > 0 else 0),
                )
            )

        return Table.create(
            reports,
            (
                r'EventName',
                r'SuccessTotal',
                r'FailureTotal',
                r'SuccessRatio',
                r'SuccessAveTime',
                r'SuccessMinTime',
                r'SuccessMaxTime',
            ),
            use_ansi=False
        )


class TaskInterface(Utils, RunnableInterface):

    def __init__(self, publisher: PublisherWithBuffer):

        self._publisher = publisher

    def success(self, name: str, resp_time: int):

        self._publisher.append(
            (name, r'success', resp_time,)
        )

    def failure(self, name: str, resp_time: int):

        self._publisher.append(
            (name, r'failure', resp_time,)
        )

    async def run(self):
        raise NotImplementedError()


class Runner(Utils, RunnableInterface):

    def __init__(self, task_cls: TaskInterface):

        global SIGNAL_PROTOCOL, SIGNAL_PORT

        self._task_cls = task_cls
        self._publisher = PublisherWithBuffer(f'{SIGNAL_PROTOCOL}://localhost:{SIGNAL_PORT}', False)

    async def run(self, times, task_num):

        self._publisher.open()

        for _ in range(times):

            tasks = MultiTasks()

            for _ in range(task_num):
                tasks.append(self._task_cls(self._publisher).run())

            await tasks

        await self._publisher.safe_close()
